# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
The classes in the astropy docs are documented by their API location,
which is not necessarily where they are defined in the source.  This
causes a problem when certain automated features of the doc build,
such as the inheritance diagrams or the `Bases` list of a class
reference a class by its canonical location rather than its "user"
location.

In the `autodoc-process-docstring` event, a mapping from the actual
name to the API name is maintained.  Later, in the `missing-reference`
event, unresolved references are looked up in this dictionary and
corrected if possible.
"""

from docutils.nodes import literal, reference

from .utils import SPHINX_LT_9

__all__ = []


def process_docstring(app, what, name, obj, options, lines):
    if isinstance(obj, type):
        env = app.env
        if not hasattr(env, 'class_name_mapping'):
            env.class_name_mapping = {}
        mapping = env.class_name_mapping
        mapping[obj.__module__ + '.' + obj.__name__] = name


def merge_mapping(app, env, docnames, env_other):
    if not hasattr(env_other, 'class_name_mapping'):
        return
    if not hasattr(env, 'class_name_mapping'):
        env.class_name_mapping = {}
    env.class_name_mapping.update(env_other.class_name_mapping)


def missing_reference_handler(app, env, node, contnode):
    """
    Handler to be connect to the sphinx 'missing-reference' event.  The handler a
    resolves reference (node) and returns a new node when sphinx could not
    originally resolve the reference.

    see `missing-reference in sphinx documentation
    <https://www.sphinx-doc.org/en/master/extdev/appapi.html#event-missing-reference>`_

    :param app: The Sphinx application object
    :param env: The build environment (``app.builder.env`)
    :param node: The ``pending_xref`` node to be resolved. Its attributes reftype,
                 reftarget, modname and classname attributes determine the type and
                 target of the reference.
    :param contnode: The node that carries the text and formatting inside the
                     future reference and should be a child of the returned
                     reference node.
    """
    # a good example of how a missing reference handle works look to
    #  https://github.com/sphinx-doc/sphinx/issues/1572#issuecomment-68590981
    #
    # Important attributes of the "node":
    #
    #      example role:  :ref:`title <target>`
    #
    #  'reftype'     - role name (in the example above 'ref' is the reftype)
    #  'reftarget'   - target of the role, as given in the role content
    #                  (in the example 'target' is the reftarget
    #  'refexplicit' - the explicit title of the role
    #                  (in the example 'title' is the refexplicit)
    #  'refdoc'      - document in which the role appeared
    #  'refdomain'   - domain of the role, in our case emtpy

    if not hasattr(env, 'class_name_mapping'):
        env.class_name_mapping = {}
    mapping = env.class_name_mapping

    reftype = node['reftype']
    reftarget = node['reftarget']
    refexplicit = node.get('refexplicit')  # default: None
    refdoc = node.get('refdoc', env.docname)
    if reftype in ('obj', 'class', 'exc', 'meth'):
        suffix = ''
        if reftarget not in mapping:
            if '.' in reftarget:
                front, suffix = reftarget.rsplit('.', 1)
            else:
                front = None
                suffix = reftarget

            if suffix.startswith('_') and not suffix.startswith('__'):
                # If this is a reference to a hidden class or method,
                # we can't link to it, but we don't want to have a
                # nitpick warning.
                return node[0].deepcopy()

            if reftype in ('obj', 'meth') and front is not None:
                if front in mapping:
                    reftarget = front
                    suffix = '.' + suffix

            if (reftype in ('class', ) and '.' in reftarget and
                    reftarget not in mapping):

                if '.' in front:
                    reftarget, _ = front.rsplit('.', 1)
                    suffix = '.' + suffix
                reftarget = reftarget + suffix
                prefix = reftarget.rsplit('.')[0]
                inventory = getattr(env, 'intersphinx_named_inventory', {})
                if (reftarget not in mapping and
                        prefix in inventory):

                    if ('py:class' in inventory[prefix] and
                            reftarget in inventory[prefix]['py:class']):
                        if SPHINX_LT_9:
                            newtarget = inventory[prefix]['py:class'][reftarget][2]
                        else:
                            newtarget = inventory[prefix]['py:class'][reftarget].uri
                        if not refexplicit and '~' not in node.rawsource:
                            contnode = literal(text=reftarget)
                        newnode = reference('', '', internal=True)
                        newnode['reftitle'] = reftarget
                        newnode['refuri'] = newtarget
                        newnode.append(contnode)

                        return newnode

        if reftarget in mapping:
            newtarget = mapping[reftarget] + suffix
            if not refexplicit and '~' not in node.rawsource:
                contnode = literal(text=newtarget)
            newnode = env.domains['py'].resolve_xref(env, refdoc, app.builder, 'class',
                                                     newtarget, node, contnode)
            if newnode is not None:
                newnode['reftitle'] = reftarget
            return newnode


def setup(app):

    app.connect('autodoc-process-docstring', process_docstring)
    app.connect('missing-reference', missing_reference_handler)
    app.connect('env-merge-info', merge_mapping)

    return {'parallel_read_safe': True,
            'parallel_write_safe': True}



__all__ = ['Foo']


class BaseFoo:
    """
    Base class for Foo
    """

    def bar(self, time):
        """
        Eat some spam in the required time.
        """
        pass


class Foo(BaseFoo):
    """
    The main foo
    """
    def hmm(self):
        pass



"""
A collection of useful functions
"""

__all__ = ['add', 'subtract', 'multiply']


def add(a, b):
    """
    Add two numbers
    """
    return a + b


def subtract(a, b):
    """
    Subtract two numbers
    """
    return a - b


def multiply(c, d):
    """
    Multiply two numbers
    """
    return c * d



__all__ = ['NonAsciiÄöüßő']


class NonAsciiÄöüßő:
    def get_äöü(self):
        """
        Return a string with common umlauts like äöüß
        """
        return 'äöü'

    def get_ß(self):
        """
        Return a string with the eszett symbol
        """
        return 'ß'



__all__ = ['Spam', 'Egg']


class BaseSpam:
    """
    Base class for Spam
    """

    def eat(self, time):
        """
        Eat some spam in the required time.
        """
        pass

    def buy(self, price):
        """
        Buy some MOAR spam.
        """
        pass


class Spam(BaseSpam):
    """
    The main spam
    """
    pass


class Egg:
    """
    An egg (no inheritance)
    """

    def eat(self, time):
        """
        Eat some egg in the required time.
        """
        pass

    def buy(self, price):
        """
        Buy some MOAR egg.
        """
        pass

    @property
    def weight(self):
        """
        The weight of an egg
        """
        return 0



"""
A module that imports objects from the standard library.
"""
from pathlib import Path
from datetime import time


__all__ = ['Path', 'time', 'add']


def add(a, b):
    """
    Add two numbers
    """
    return a + b



# namedtuple is needed for find_mod_objs so it can have a non-local module

from collections import namedtuple

from ..utils import find_mod_objs


def test_find_mod_objs():
    lnms, fqns, objs = find_mod_objs('sphinx_automodapi')

    # just check for astropy.test ... other things might be added, so we
    # shouldn't check that it's the only thing
    assert lnms == []

    lnms, fqns, objs = find_mod_objs(
        'sphinx_automodapi.tests.test_utils', onlylocals=False)

    assert namedtuple in objs

    lnms, fqns, objs = find_mod_objs(
        'sphinx_automodapi.tests.test_utils', onlylocals=True)
    assert 'namedtuple' not in lnms
    assert 'collections.namedtuple' not in fqns
    assert namedtuple not in objs


def test_find_mod_objs_with_list_of_modules():
    lnms, fqns, objs = find_mod_objs(
        'sphinx_automodapi.tests.test_utils', onlylocals=['sphinx_automodapi'])

    assert namedtuple not in objs
    assert find_mod_objs in objs

    lnms, fqns, objs = find_mod_objs(
        'sphinx_automodapi.tests.test_utils', onlylocals=['collections'])

    assert namedtuple in objs
    assert find_mod_objs not in objs

    lnms, fqns, objs = find_mod_objs(
        'sphinx_automodapi.tests.test_utils', onlylocals=['collections',
                                                          'sphinx_automodapi'])

    assert namedtuple in objs
    assert find_mod_objs in objs



class ClassWithAttribute:
    """A class with an attribute."""
    my_attribute = 1

    def my_method(self):
        """A method."""
        pass

    @property
    def my_property(self):
        """A property."""
        return 1



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
from copy import deepcopy

from sphinx.cmd.build import build_main

from . import cython_testpackage  # noqa

__all__ = ['write_conf', 'run_sphinx_in_tmpdir']


intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None)
}

DEFAULT_CONF = {'source_suffix': '.rst',
                'master_doc': 'index',
                'nitpicky': True,
                'extensions': ['sphinx.ext.intersphinx', 'sphinx_automodapi.automodapi'],
                'suppress_warnings': ['app.add_directive', 'app.add_node'],
                'intersphinx_mapping': intersphinx_mapping,
                'automodapi_toctreedirnm': 'api',
                'automodapi_writereprocessed': True,
                'automodapi_inheritance_diagram': True,
                'automodsumm_writereprocessed': True}


def write_conf(filename, conf):
    with open(filename, 'w') as f:
        for key, value in conf.items():
            f.write(f"{key} = {repr(conf[key])}\n")


def run_sphinx_in_tmpdir(tmpdir, additional_conf={}, expect_error=False):

    start_dir = os.path.abspath('.')

    conf = deepcopy(DEFAULT_CONF)
    conf.update(additional_conf)

    write_conf(tmpdir.join('conf.py').strpath, conf)

    argv = ['-W', '-b', 'html', '.', '_build/html']

    try:
        os.chdir(tmpdir.strpath)
        status = build_main(argv=argv)
    finally:
        os.chdir(start_dir)

    if expect_error:
        assert status != 0
    else:
        assert status == 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from copy import copy

import pytest
from docutils.parsers.rst import directives, roles

from . import cython_testpackage  # noqa
from .helpers import run_sphinx_in_tmpdir


def setup_function(func):
    # This can be replaced with the docutils_namespace context manager once
    # it is in a stable release of Sphinx
    func._directives = copy(directives._directives)
    func._roles = copy(roles._roles)


def teardown_function(func):
    directives._directives = func._directives
    roles._roles = func._roles


am_replacer_str = """
This comes before

.. automodapi:: sphinx_automodapi.tests.example_module.mixed
{options}

This comes after
"""

am_replacer_basic_expected = """
This comes before


sphinx_automodapi.tests.example_module.mixed Module
---------------------------------------------------

.. automodule:: sphinx_automodapi.tests.example_module.mixed

Functions
^^^^^^^^^

.. automodsumm:: sphinx_automodapi.tests.example_module.mixed
    :functions-only:
    :toctree: api

Classes
^^^^^^^

.. automodsumm:: sphinx_automodapi.tests.example_module.mixed
    :classes-only:
    :toctree: api

Class Inheritance Diagram
^^^^^^^^^^^^^^^^^^^^^^^^^

.. automod-diagram:: sphinx_automodapi.tests.example_module.mixed
    :private-bases:
    :parts: 1

This comes after
"""


def test_am_replacer_basic(tmpdir):
    """
    Tests replacing an ".. automodapi::" with the automodapi no-option
    template
    """

    with open(tmpdir.join('index.rst').strpath, 'w') as f:
        f.write(am_replacer_str.format(options=''))

    run_sphinx_in_tmpdir(tmpdir)

    with open(tmpdir.join('index.rst.automodapi').strpath) as f:
        result = f.read()

    assert result == am_replacer_basic_expected


am_replacer_repr_str = """
This comes before with spéciàl çhars

.. automodapi:: sphinx_automodapi.tests.example_module.mixed
{options}

This comes after
"""


@pytest.mark.parametrize('writereprocessed', [False, True])
def test_am_replacer_writereprocessed(tmpdir, writereprocessed):
    """
    Tests the automodapi_writereprocessed option
    """

    with open(tmpdir.join('index.rst').strpath, 'w', encoding='utf-8') as f:
        f.write(am_replacer_repr_str.format(options=''))

    run_sphinx_in_tmpdir(tmpdir, additional_conf={'automodapi_writereprocessed': writereprocessed})

    assert tmpdir.join('index.rst.automodapi').isfile() is writereprocessed


am_replacer_noinh_expected = """
This comes before


sphinx_automodapi.tests.example_module.mixed Module
---------------------------------------------------

.. automodule:: sphinx_automodapi.tests.example_module.mixed

Functions
^^^^^^^^^

.. automodsumm:: sphinx_automodapi.tests.example_module.mixed
    :functions-only:
    :toctree: api

Classes
^^^^^^^

.. automodsumm:: sphinx_automodapi.tests.example_module.mixed
    :classes-only:
    :toctree: api


This comes after
""".format(empty='')


def test_am_replacer_noinh(tmpdir):
    """
    Tests replacing an ".. automodapi::" with no-inheritance-diagram
    option
    """

    ops = ['', ':no-inheritance-diagram:']
    ostr = '\n    '.join(ops)

    with open(tmpdir.join('index.rst').strpath, 'w') as f:
        f.write(am_replacer_str.format(options=ostr))

    run_sphinx_in_tmpdir(tmpdir)

    with open(tmpdir.join('index.rst.automodapi').strpath) as f:
        result = f.read()

    assert result == am_replacer_noinh_expected


am_replacer_titleandhdrs_expected = """
This comes before


sphinx_automodapi.tests.example_module.mixed Module
&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&

.. automodule:: sphinx_automodapi.tests.example_module.mixed

Functions
*********

.. automodsumm:: sphinx_automodapi.tests.example_module.mixed
    :functions-only:
    :toctree: api

Classes
*******

.. automodsumm:: sphinx_automodapi.tests.example_module.mixed
    :classes-only:
    :toctree: api

Class Inheritance Diagram
*************************

.. automod-diagram:: sphinx_automodapi.tests.example_module.mixed
    :private-bases:
    :parts: 1


This comes after
"""


def test_am_replacer_titleandhdrs(tmpdir):
    """
    Tests replacing an ".. automodapi::" entry with title-setting and header
    character options.
    """

    ops = ['', ':headings: &*']
    ostr = '\n    '.join(ops)

    with open(tmpdir.join('index.rst').strpath, 'w') as f:
        f.write(am_replacer_str.format(options=ostr))

    run_sphinx_in_tmpdir(tmpdir)

    with open(tmpdir.join('index.rst.automodapi').strpath) as f:
        result = f.read()

    assert result == am_replacer_titleandhdrs_expected


def test_am_replacer_titleandhdrs_invalid(tmpdir, capsys):
    """
    Tests replacing an ".. automodapi::" entry with title-setting and header
    character options.
    """

    ops = ['', ':headings: &']
    ostr = '\n    '.join(ops)

    with open(tmpdir.join('index.rst').strpath, 'w') as f:
        f.write(am_replacer_str.format(options=ostr))

    run_sphinx_in_tmpdir(tmpdir, expect_error=True)

    stdout, stderr = capsys.readouterr()
    assert "Not enough headings (got 1, need 2), using default -^" in stderr


am_replacer_nomain_str = """
This comes before

.. automodapi:: sphinx_automodapi.tests.example_module.functions
    :no-main-docstr:

This comes after
"""

am_replacer_nomain_expected = """
This comes before


sphinx_automodapi.tests.example_module.functions Module
-------------------------------------------------------



Functions
^^^^^^^^^

.. automodsumm:: sphinx_automodapi.tests.example_module.functions
    :functions-only:
    :toctree: api


This comes after
""".format(empty='')


def test_am_replacer_nomain(tmpdir):
    """
    Tests replacing an ".. automodapi::" with "no-main-docstring" .
    """

    with open(tmpdir.join('index.rst').strpath, 'w') as f:
        f.write(am_replacer_nomain_str.format(options=''))

    run_sphinx_in_tmpdir(tmpdir)

    with open(tmpdir.join('index.rst.automodapi').strpath) as f:
        result = f.read()

    assert result == am_replacer_nomain_expected


am_replacer_skip_str = """
This comes before

.. automodapi:: sphinx_automodapi.tests.example_module.functions
    :skip: add
    :skip: subtract

This comes after
"""

am_replacer_skip_expected = """
This comes before


sphinx_automodapi.tests.example_module.functions Module
-------------------------------------------------------

.. automodule:: sphinx_automodapi.tests.example_module.functions

Functions
^^^^^^^^^

.. automodsumm:: sphinx_automodapi.tests.example_module.functions
    :functions-only:
    :toctree: api
    :skip: add,subtract


This comes after
""".format(empty='')


def test_am_replacer_skip(tmpdir):
    """
    Tests using the ":skip: option in an ".. automodapi::" .
    """

    with open(tmpdir.join('index.rst').strpath, 'w') as f:
        f.write(am_replacer_skip_str.format(options=''))

    run_sphinx_in_tmpdir(tmpdir)

    with open(tmpdir.join('index.rst.automodapi').strpath) as f:
        result = f.read()

    assert result == am_replacer_skip_expected


am_replacer_skip_stdlib_str = """
This comes before

.. automodapi:: sphinx_automodapi.tests.example_module.stdlib
    :skip: time
    :skip: Path

This comes after
"""


am_replacer_skip_stdlib_expected = """
This comes before


sphinx_automodapi.tests.example_module.stdlib Module
----------------------------------------------------

.. automodule:: sphinx_automodapi.tests.example_module.stdlib

Functions
^^^^^^^^^

.. automodsumm:: sphinx_automodapi.tests.example_module.stdlib
    :functions-only:
    :toctree: api
    :skip: time,Path


This comes after
""".format(empty='')


def test_am_replacer_skip_stdlib(tmpdir):
    """
    Tests using the ":skip:" option in an ".. automodapi::"
    that skips objects imported from the standard library.
    This is a regression test for #141
    """

    with open(tmpdir.join('index.rst').strpath, 'w') as f:
        f.write(am_replacer_skip_stdlib_str.format(options=''))

    run_sphinx_in_tmpdir(tmpdir)

    with open(tmpdir.join('index.rst.automodapi').strpath) as f:
        result = f.read()

    assert result == am_replacer_skip_stdlib_expected


am_replacer_include_stdlib_str = """
This comes before

.. automodapi:: sphinx_automodapi.tests.example_module.stdlib
    :include: add
    :allowed-package-names: pathlib, datetime, sphinx_automodapi

This comes after
"""

am_replacer_include_stdlib_expected = """
This comes before


sphinx_automodapi.tests.example_module.stdlib Module
----------------------------------------------------

.. automodule:: sphinx_automodapi.tests.example_module.stdlib

Functions
^^^^^^^^^

.. automodsumm:: sphinx_automodapi.tests.example_module.stdlib
    :functions-only:
    :toctree: api
    :skip: Path,time
    :allowed-package-names: pathlib,datetime,sphinx_automodapi


This comes after
""".format(empty='')


def test_am_replacer_include_stdlib(tmpdir):
    """
    Tests using the ":include: option in an ".. automodapi::"
    in the presence of objects imported from the standard library.
    """

    with open(tmpdir.join('index.rst').strpath, 'w') as f:
        f.write(am_replacer_include_stdlib_str.format(options=''))

    run_sphinx_in_tmpdir(tmpdir)

    with open(tmpdir.join('index.rst.automodapi').strpath) as f:
        result = f.read()

    assert result == am_replacer_include_stdlib_expected


am_replacer_include_str = """
This comes before

.. automodapi:: sphinx_automodapi.tests.example_module.functions
    :include: add
    :include: subtract

This comes after
"""

am_replacer_include_expected = """
This comes before


sphinx_automodapi.tests.example_module.functions Module
-------------------------------------------------------

.. automodule:: sphinx_automodapi.tests.example_module.functions

Functions
^^^^^^^^^

.. automodsumm:: sphinx_automodapi.tests.example_module.functions
    :functions-only:
    :toctree: api
    :skip: multiply


This comes after
""".format(empty='')


def test_am_replacer_include(tmpdir):
    """
    Tests using the ":include: option in an ".. automodapi::" .
    """

    with open(tmpdir.join('index.rst').strpath, 'w') as f:
        f.write(am_replacer_include_str.format(options=''))

    run_sphinx_in_tmpdir(tmpdir)

    with open(tmpdir.join('index.rst.automodapi').strpath) as f:
        result = f.read()

    assert result == am_replacer_include_expected


am_replacer_invalidop_str = """
This comes before

.. automodapi:: sphinx_automodapi.tests.example_module.functions
    :invalid-option:

This comes after
"""


def test_am_replacer_invalidop(tmpdir, capsys):
    """
    Tests that a sphinx warning is produced with an invalid option.
    """

    with open(tmpdir.join('index.rst').strpath, 'w') as f:
        f.write(am_replacer_invalidop_str.format(options=''))

    run_sphinx_in_tmpdir(tmpdir, expect_error=True)

    stdout, stderr = capsys.readouterr()
    assert "Found additional options invalid-option in automodapi." in stderr


am_replacer_cython_str = """
This comes before

.. automodapi:: apyhtest_eva.unit02
{options}

This comes after
"""

am_replacer_cython_expected = """
This comes before


apyhtest_eva.unit02 Module
--------------------------

.. automodule:: apyhtest_eva.unit02

Functions
^^^^^^^^^

.. automodsumm:: apyhtest_eva.unit02
    :functions-only:
    :toctree: api

This comes after
""".format(empty='')


def test_am_replacer_cython(tmpdir, cython_testpackage):  # noqa
    """
    Tests replacing an ".. automodapi::" for a Cython module.
    """

    with open(tmpdir.join('index.rst').strpath, 'w') as f:
        f.write(am_replacer_cython_str.format(options=''))

    run_sphinx_in_tmpdir(tmpdir)

    with open(tmpdir.join('index.rst.automodapi').strpath) as f:
        result = f.read()

    assert result == am_replacer_cython_expected



"""Test classes containing __slots__

Instance attributes named in ``__slots__`` can be introspected and are listed
in the Attributes section of the class documentation. Class attributes are
listed in the same section of the generated docs so docstrings should be used
to distinguish class attributes vs instance attributes. Regular instance
attributes are dynamically inserted into ``__dict__`` and cannot be reliably
introspected so they're not included in the documentation.
"""
from __future__ import annotations

__all__ = ['SlotDict', 'DerivedParam', 'DerivedSlotParam',]


class SlotDict:
    """
    A class that uses __slots__ and __dict__ for its attribute namespace.
    """
    __slots__ = {
        "instance_attr": "instance attribute docstring can be added here",
        "__dict__": None,   # Allows additional instance attributes to be added
    }

    class_attr = "class attribute value"
    """(class attr) this is a class attribute."""

    def __init__(self, param: str, other_param: str):
        """
        Initializes a SlotDict object.

        Parameters
        ----------
        param : str
            A parameter
        other_param : str
            Another parameter
        """

        self.instance_attr = param
        """Instance attributes declared in slots can also define their docstring
        here
        """

        if other_param is not None:
            self.other_attr = other_param
            """This instance attribute is dynamic (not declared in a slot) so
            it's not included in the docs
            """

    def my_method(self):
        """
        Prints the SlotDict parameters.
        """
        print(f"instance_attr: {self.instance_attr}")
        print(f"other_attr: {self.other_attr}")


class DerivedParam(SlotDict):
    """
    Extends SlotDict by adding an extra parameter
    """
    def __init__(self, param: str, other_param: str, extra_param: str):
        """
        Initializes a DerivedParam object.

        Parameters
        ----------
        param : str
            A parameter
        other_param : str
            Another parameter
        extra_param : str
            An extra parameter
        """
        super().__init__(param, other_param)
        self.extra_attr = extra_param

    def derived_from_slot_class_method(self):
        """
        Prints the DerivedParam parameters.
        """
        print(f"instance_attr: {self.instance_attr}")
        print(f"other_attr: {self.other_attr}")
        print(f"extra_attr: {self.extra_attr}")


class DerivedSlotParam(SlotDict):
    """
    Extends SlotDict by adding a slot parameter
    """

    __slots__ = ('extra_attr',)

    def __init__(self, param: str, other_param: str, extra_param: str):
        """
        Initializes a DerivedSlotParam object.

        Parameters
        ----------
        param : str
            A parameter
        other_param : str
            Another parameter
        extra_param : str
            An extra parameter
        """
        super().__init__(param, other_param)
        self.extra_attr = extra_param

    def derived_from_slot_class_method(self):
        """
        Prints the DerivedSlotParam parameters.
        """
        print(f"instance_attr: {self.instance_attr}")
        print(f"other_attr: {self.other_attr}")
        print(f"extra_attr: {self.extra_attr}")



"""
A collection of useful classes and functions
"""

__all__ = ['add', 'MixedSpam']


def add(a, b):
    """
    Add two numbers
    """
    return a + b


class MixedSpam:
    """
    Special spam
    """

    def eat(self, time):
        """
        Eat special spam in the required time.
        """
        pass
